#!/usr/bin/env bash

export KEY=Array1

declare -a subkeys
declare -i i

s='{
    "Array1": [
	    {
		"key1": "some string",
		"key2": "another string",
		"Array2": [
		    {
			"key3": "int",
			"Array3": [],
			"Array4": [1,2],
			"Array5": [
			    {
				"key4": "one more string",
				"key5": "last string"
			    }
			]
		    }
		]
	    }
	]
}
'

echo Test:
echo jq 'if [type, length] != ["object",1] then error(...'
jq.sh 'if [type, length] != ["object",1] then error("The input is not a object with 1 element")
elif keys[] != "'$KEY'" then error("The input object have not key '$KEY'")
elif [.[env.KEY]|type, length] != ["array",1] then error("The input have not subarray with 1 element")
elif .[env.KEY][0]|type != "object" then error("The input is not {\"'$KEY'\":[{...}]}")
end' <<< "$s" || exit $?
KEY_VAL="$(jq.sh ".$KEY[0]" <<< "$s")"
subkeys=($(jq.sh 'keys_unsorted[]' <<< "$KEY_VAL"))
OKD=('"some string"' '"another string"'
 '[{"key3":"int","Array3":[],"Array4":[1,2],"Array5":[{"key4":"one more string","key5":"last string"}]}]')
for ((i=0;i<${#subkeys[*]};i++)); do
	O=$(jq.sh -c ".${subkeys[i]}" <<< "$KEY_VAL")
	if [[ ${OKD[i]} != "$O" ]]; then
		echo "expected '${OKD[i]}', but returned '$O'" >&2
		exit 1
	fi
done
echo "Ok"

echo Test:
echo jq '(2,keys|length, type),keys,{a:0, a:1, (.x):(.Y+2)}' '<<<' '{"x":"s", "Y":42, "z":null}'
Ok="$(jq -c '(2,keys|length, type),keys,{a:0, a:1, (.x):(.Y+2)}' <<< '{"x":"s", "Y":42, "z":null}')"
O="$(jq.sh -c '(2,keys|length, type),keys,{a:0, a:1, (.x):(.Y+2)}' <<< '{"x":"s", "Y":42, "z":null}')"
if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"

Ok='"a2_e1_v"
"a2_e2_v"
"a3_e1_v"
"a3_e2_v"'
echo Test:
echo jq.sh '.[][]' '<<<' '{"a1_e1":{"a2_e1":"a2_e1_v", "a2_e2":"a2_e2_v"}, "a1_e2":{"a3_e1":"a3_e1_v", "a3_e2":"a3_e2_v"}}'
 O=$(jq.sh '.[][]'  <<<  '{"a1_e1":{"a2_e1":"a2_e1_v", "a2_e2":"a2_e2_v"}, "a1_e2":{"a3_e1":"a3_e1_v", "a3_e2":"a3_e2_v"}}')

if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"

Ok='[1,2,[],[null],[null,3],[0,null],[1,null]]'
echo Test:
echo jq.sh 'sort' '<<<' '[2,1,[1,null],[0,null],[null,3],[],[null]]'
 O=$(jq.sh -c 'sort'  <<<  '[2,1,[1,null],[0,null],[null,3],[],[null]]')
if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"

Ok='[{"X":"s"},{"a":null,"x":null},{"a":2,"x":null},{"a":3},{"x":null}]'
echo Test:
echo jq.sh 'sort' '<<<' '[{a:2,x:null},{a:3},{x:null},{a:null,x:null},{X:s}]'
 O=$(jq.sh -c 'sort'  <<<  '[{a:2,x:null},{a:3},{x:null},{a:null,x:null},{X:s}]')
if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"

Ok='[10,5]'
echo Test:
echo jq.sh -c '. as $i|[(.*2|. as $i| $i), $i]' '<<<' '5'
 O=$(jq.sh -c '. as $i|[(.*2|. as $i| $i), $i]' <<< '5')
if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"

Ok='{"title":"Frist psot","author":"Anonymous Coward"}
{"title":"A well-written article","author":"Person McPherson"}'
echo Test:
echo jq.sh -c '.realnames as $names | (.posts[] | {title, author: $names[.author]})'
 O=$(jq.sh -c '.realnames as $names | (.posts[] | {title, author: $names[.author]})' <<< '{
 "posts": [{"title": "Frist psot", "author": "anon"},
	   {"title": "A well-written article", "author": "person1"}],
 "realnames": {"anon": "Anonymous Coward", "person1": "Person McPherson"}
}')
if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"

s='{
  "glossary": {
    "title": "example glossary",
    "GlossDiv": {
      "title": "S",
      "GlossList": {
	"GlossEntry": {
	  "ID": "SGML",
	  "SortAs": "SGML",
	  "GlossTerm": "Standard Generalized Markup Language",
	  "Acronym": "SGML",
	  "Abbrev": "ISO 8879:1986",
	  "GlossDef": {
	    "para": "A meta-markup language, used to create markup languages such as DocBook.",
	    "GlossSeeAlso": [
	      "GML",
	      "XML"
	    ]
	  },
	  "GlossSee": "markup"
	}
      }
    }
  }
}
{
  "menu": {
    "id": "file",
    "value": "File",
    "popup": {
      "menuitem": [
	{
	  "value": "New",
	  "onclick": "CreateNewDoc()"
	},
	{
	  "value": "Open",
	  "onclick": "OpenDoc()"
	},
	{
	  "value": "Close",
	  "onclick": "CloseDoc()"
	}
      ]
    }
  }
}
{
  "widget": {
    "debug": "on",
    "window": {
      "title": "Sample Konfabulator Widget",
      "name": "main_window",
      "width": 500,
      "height": 500
    },
    "image": {
      "src": "Images/Sun.png",
      "name": "sun1",
      "hOffset": 250,
      "vOffset": 250,
      "alignment": "center"
    },
    "text": {
      "data": "Click Here",
      "size": 36,
      "style": "bold",
      "name": "text1",
      "hOffset": 250,
      "vOffset": 100,
      "alignment": "center",
      "onMouseUp": "sun1.opacity = (sun1.opacity / 100) * 90;"
    }
  }
}
{
  "web-app": {
    "servlet": [
      {
	"servlet-name": "cofaxCDS",
	"servlet-class": "org.cofax.cds.CDSServlet",
	"init-param": {
	  "configGlossary:installationAt": "Philadelphia, PA",
	  "configGlossary:adminEmail": "ksm@pobox.com",
	  "configGlossary:poweredBy": "Cofax",
	  "configGlossary:poweredByIcon": "/images/cofax.gif",
	  "configGlossary:staticPath": "/content/static",
	  "templateProcessorClass": "org.cofax.WysiwygTemplate",
	  "templateLoaderClass": "org.cofax.FilesTemplateLoader",
	  "templatePath": "templates",
	  "templateOverridePath": "",
	  "defaultListTemplate": "listTemplate.htm",
	  "defaultFileTemplate": "articleTemplate.htm",
	  "useJSP": false,
	  "jspListTemplate": "listTemplate.jsp",
	  "jspFileTemplate": "articleTemplate.jsp",
	  "cachePackageTagsTrack": 200,
	  "cachePackageTagsStore": 200,
	  "cachePackageTagsRefresh": 60,
	  "cacheTemplatesTrack": 100,
	  "cacheTemplatesStore": 50,
	  "cacheTemplatesRefresh": 15,
	  "cachePagesTrack": 200,
	  "cachePagesStore": 100,
	  "cachePagesRefresh": 10,
	  "cachePagesDirtyRead": 10,
	  "searchEngineListTemplate": "forSearchEnginesList.htm",
	  "searchEngineFileTemplate": "forSearchEngines.htm",
	  "searchEngineRobotsDb": "WEB-INF/robots.db",
	  "useDataStore": true,
	  "dataStoreClass": "org.cofax.SqlDataStore",
	  "redirectionClass": "org.cofax.SqlRedirection",
	  "dataStoreName": "cofax",
	  "dataStoreDriver": "com.microsoft.jdbc.sqlserver.SQLServerDriver",
	  "dataStoreUrl": "jdbc:microsoft:sqlserver://LOCALHOST:1433;DatabaseName=goon",
	  "dataStoreUser": "sa",
	  "dataStorePassword": "dataStoreTestQuery",
	  "dataStoreTestQuery": "SET NOCOUNT ON;select test='"'test'"';",
	  "dataStoreLogFile": "/usr/local/tomcat/logs/datastore.log",
	  "dataStoreInitConns": 10,
	  "dataStoreMaxConns": 100,
	  "dataStoreConnUsageLimit": 100,
	  "dataStoreLogLevel": "debug",
	  "maxUrlLength": 500
	}
      },
      {
	"servlet-name": "cofaxEmail",
	"servlet-class": "org.cofax.cds.EmailServlet",
	"init-param": {
	  "mailHost": "mail1",
	  "mailHostOverride": "mail2"
	}
      },
      {
	"servlet-name": "cofaxAdmin",
	"servlet-class": "org.cofax.cds.AdminServlet"
      },
      {
	"servlet-name": "fileServlet",
	"servlet-class": "org.cofax.cds.FileServlet"
      },
      {
	"servlet-name": "cofaxTools",
	"servlet-class": "org.cofax.cms.CofaxToolsServlet",
	"init-param": {
	  "templatePath": "toolstemplates/",
	  "log": 1,
	  "logLocation": "/usr/local/tomcat/logs/CofaxTools.log",
	  "logMaxSize": "",
	  "dataLog": 1,
	  "dataLogLocation": "/usr/local/tomcat/logs/dataLog.log",
	  "dataLogMaxSize": "",
	  "removePageCache": "/content/admin/remove?cache=pages&id=",
	  "removeTemplateCache": "/content/admin/remove?cache=templates&id=",
	  "fileTransferFolder": "/usr/local/tomcat/webapps/content/fileTransferFolder",
	  "lookInContext": 1,
	  "adminGroupID": 4,
	  "betaServer": true
	}
      }
    ],
    "servlet-mapping": {
      "cofaxCDS": "/",
      "cofaxEmail": "/cofaxutil/aemail/*",
      "cofaxAdmin": "/admin/*",
      "fileServlet": "/static/*",
      "cofaxTools": "/tools/*"
    },
    "taglib": {
      "taglib-uri": "cofax.tld",
      "taglib-location": "/WEB-INF/tlds/cofax.tld"
    }
  }
}
{
  "menu": {
    "header": "SVG Viewer",
    "items": [
      {
	"id": "Open"
      },
      {
	"id": "OpenNew",
	"label": "Open New"
      },
      null,
      {
	"id": "ZoomIn",
	"label": "Zoom In"
      },
      {
	"id": "ZoomOut",
	"label": "Zoom Out"
      },
      {
	"id": "OriginalView",
	"label": "Original View"
      },
      null,
      {
	"id": "Quality"
      },
      {
	"id": "Pause"
      },
      {
	"id": "Mute"
      },
      null,
      {
	"id": "Find",
	"label": "Find..."
      },
      {
	"id": "FindAgain",
	"label": "Find Again"
      },
      {
	"id": "Copy"
      },
      {
	"id": "CopyAgain",
	"label": "Copy Again"
      },
      {
	"id": "CopySVG",
	"label": "Copy SVG"
      },
      {
	"id": "ViewSVG",
	"label": "View SVG"
      },
      {
	"id": "ViewSource",
	"label": "View Source"
      },
      {
	"id": "SaveAs",
	"label": "Save As"
      },
      null,
      {
	"id": "Help"
      },
      {
	"id": "About",
	"label": "About Adobe CVG Viewer..."
      }
    ]
  }
}'
echo Test:
echo jq . exampl.txt
O=$(jq.sh . <<< "$s")
s=$(expand <<< "$s")
if [[ $O != "$s" ]]; then
	echo "exampl.txt uncompared" >&2
	exit 1
fi
echo "Ok"

echo Test:
echo jq -n -c '"inter\("pol \("x"+"y ")" + "ation")"'
Ok="$(jq -n -c '"inter\("pol \("x"+"y ")" + "ation")"')"
O="$(jq.sh -n -c '"inter\("pol \("x"+"y ")" + "ation")"')"
if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"

echo Test:
echo jq -n -c 'unique' '<<<' '[3,4,2,5,1,4,5,null,true,false,1.1]'
Ok='[null,false,true,1,1.1,2,3,4,5]'
O="$(jq.sh -c 'unique' <<< '[3,4,2,5,1,4,5,null,true,false,1.1]')"
if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"

echo Test:
echo jq -c 'del(..|select(length==3))' '<<<' '["foo", "foo2", "bar", "x"]'
Ok="$(jq -c 'del(..|select(length==3))' <<< '["foo", "foo2", "bar", "x"]')"
O="$(jq.sh -c 'del(..|select(length==3))' <<< '["foo", "foo2", "bar", "x"]')"
if [[ $O != "$Ok" ]]; then
	echo "expected '$Ok', but returned '$O'" >&2
	exit 1
fi
echo "Ok"
