/*
 * Revision 1.2  86/12/27  17:10:35  avg
 * Based by BSD 2.9 and MNOS RL 1.2.
 *
 * yacc for bash
 * Copyright (c) 1986-2019 may safely be consumed by a BSD or GPL license.
 * Written by:   Vladimir Oleynik <dzo@simtreas.ru>
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

#define ediag(e,r) (e)

/*  MANIFEST CONSTANT DEFINITIONS */

/* base of nonterminal internal numbers */
#define NTBASE 010000

/* internal codes for error and accept actions */
#define ERRCODE         8190
#define ACCEPTCODE      8191

/* sizes and limits */

/* relationships which must hold:
TBITSET ints must hold NTERMS+1 bits...
WSETSIZE >= NNONTERM
LSETSIZE >= NNONTERM
TEMPSIZE >= NTERMS + NNONTERMs + 1
TEMPSIZE >= NSTATES
*/

#define ACTSIZE 12000
#define MEMSIZE 12000
#define NSTATES 750
#define NTERMS 127
#define NPROD 600
#define NNONTERM 300
#define TEMPSIZE 1200
#define CNAMSZ 5000
#define LSETSIZE 600
#define WSETSIZE 350

#define NAMESIZE 50

#if 1
/* int WORD32 */
# define TBITSET ((32+NTERMS)/32)

	/* bit packing macros (may be machine dependent) */
# define BIT(a,i) ((a)[(i)>>5] & (1<<((i)&037)))
# define SETBIT(a,i) ((a)[(i)>>5] |= (1<<((i)&037)))

	/* number of words needed to hold n+1 bits */
# define NWORDS(n) (((n)+32)/32)

#else

# define TBITSET ((16+NTERMS)/16)

	/* bit packing macros (may be machine dependent) */
# define BIT(a,i) ((a)[(i)>>4] & (1<<((i)&017)))
# define SETBIT(a,i) ((a)[(i)>>4] |= (1<<((i)&017)))

	/* number of words needed to hold n+1 bits */
# define NWORDS(n) (((n)+16)/16)
#endif

#if WSETSIZE < NNONTERM || LSETSIZE < NNONTERM || TEMPSIZE < NTERMS + NNONTERM + 1 || TEMPSIZE < NSTATES
# include "     "
#endif

/* associativities */
#define NOASC 0 /* no assoc. */
#define LASC 1  /* left assoc. */
#define RASC 2  /* right assoc. */
#define BASC 3  /* binary assoc. */

/* flags for state generation */
#define DONE 0
#define MUSTDO 1
#define MUSTLOOKAHEAD 2

/* flags for a rule having an action, and being reduced */

#define ACTFLAG 04
#define REDFLAG 010

/* output parser flags */
#define YYFLAG1 (-1000)

/* macros for getting associativity and precedence levels */

#define ASSOC(i) ((i)&03)
#define PLEVEL(i) (((i)>>4)&077)
#define TYPE(i)  ((i>>10)&077)

/* macros for setting associativity and precedence levels */

#define SETASC(i,j)  i |= j
#define SETPLEV(i,j) i |= (j<<4)
#define SETTYPE(i,j) i |= (j<<10)

/* looping macros */

#define TLOOP(i) for(i=1;i<=ntokens;++i)
#define NTLOOP(i) for(i=0;i<=nnonter;++i)
#define PLOOP(s,i) for(i=s;i<nprod;++i)
#define SLOOP(i) for(i=0;i<nstate;++i)
#define WSBUMP(x) ++x
#define WSLOOP(s,j) for(j=s;j<cwp;++j)
#define ITMLOOP(i,p,q) q=pstate[i+1];for(p=pstate[i];p<q;++p)
#define SETLOOP(i) for(i=0;i<tbitset;++i)

/* input and output file descriptors */
static FILE * finput;           /* yacc input file */
static FILE * faction;          /* file for saving actions */
static FILE * ftable;           /* OFILE */
static FILE * ftemp;            /* tempfile to pass 2 */
static FILE * foutput;          /* y_output file */
static FILE * Y2_MARK_fi;       /* input file after %% */


/* structure declarations */

struct looksets {
	int lset[TBITSET];
};

struct item {
	int *pitem;
	struct looksets *look;
};

struct toksymb {
	char *name;
	int value;
};

struct ntsymb {
	char *name;
	int tvalue;
};

struct wset {
	int *pitem;
	int flag;
	struct looksets ws;
};

/* communication variables between various I/O routines */
static char *infile;   /* input file name */
static int numbval;    /* value of an input number */
static char tokname[NAMESIZE]; /* input token name */

/* storage of names */
static char *cnames;  /* place where token and nonterminal names are stored */
static int cnamsz = CNAMSZ;    /* size of cnames */
static char * cnamp;  /* place where next name is to be put in */
static int ndefout = 3;  /* number of defined symbols output */

/* symbol tables for tokens and nonterminals */
static int ntokens = 0;                /* number of tokens */
static struct toksymb tokset[NTERMS];
static int toklev[NTERMS];             /* vector with the precedence of the terminals */
static int nnonter = -1;               /* the number of nonterminals */
static struct ntsymb nontrst[NNONTERM];
static int start;      /* start symbol */

/* assigned token type values */
static int extval = 0;


/* storage for grammar rules */
static int *mem0;              /* production storage */
static int *mem;
static int nprod= 1;           /* number of productions */
static int *prdptr[NPROD];     /* pointers to descriptions of productions */
static int levprd[NPROD] ;     /* precedence levels for the productions */

/* lookahead computations */
static int tbitset;  /* size of lookahead sets */
static struct looksets lkst [ LSETSIZE ];
static int nlset = 0; /* next lookahead set index */
static int nolook = 0; /* flag to suppress lookahead computations */
static struct looksets clset;  /* temporary storage for lookahead computations */

/* working set computations */
static struct wset wsets[ WSETSIZE ];
static struct wset *cwp;

/* state information */
static int nstate = 0;                 /* number of states */
static struct item *pstate[NSTATES+2]; /* pointers to the descriptions of the states */
static int tystate[NSTATES];           /* contains type information about the states */
static int indgo[NSTATES];             /* index to the stored goto table */
static int tstates[ NTERMS ];          /* states generated by terminal gotos */
static int ntstates[ NNONTERM ];       /* states generated by nonterminal gotos */
static int mstates[ NSTATES ];         /* chain of overflows of term/nonterm generation lists  */

/* storage for the actions in the parser */
static int  *amem;                     /* action table storage */ /* 31.03.85 @VG */
static int  *memp;                     /* next free action table position */
static int * zzmemsz;

/* other storage areas */
static int *temp1;             /* temporary storage, indexed by terms + ntokens or states */
static int lineno = 1;         /* current input line number */
static int fatfl = 1;          /* if on, error is fatal */
static int nerrors = 0;        /* number of errors */

/* storage for information about the nonterminals */
static int **pres[NNONTERM+2];  /* vector of pointers to productions yielding each nonterminal */
static struct looksets *pfirst[NNONTERM+2];  /* vector of pointers to first sets for each nonterminal */
static int pempty[NNONTERM+1];  /* vector of nonterminals nontrivially deriving e */


static int lastred;            /* the number of the last reduction of a state */
static int defact[NSTATES];    /* the default actions of states */

/* statistics collection variables */
static struct wset *zzcwp = wsets;
static int zzgoent = 0;
static int zzgobest = 0;
static int zzacent = 0;
static int zzexcp = 0;
static int zzclose = 0;
static int zzsrconf = 0;
static int zzrrconf = 0;

static int * ggreed = lkst[0].lset;
static int * pgo = wsets[0].ws.lset;
static int *yypgo = &nontrst[0].tvalue;

static int maxspr = 0;  /* maximum spread of any entry */
static int maxoff = 0;  /* maximum offset into a array */
static int *pmem;
static int *maxa;
# define Y4_NOMORE -1000

static const char *YYSTYPE = "";

/* default settings for a number of macros */

/* name of yacc tempfiles */

#ifndef TEMPNAME
# define TEMPNAME "yacc.tmp"
#endif

#ifndef ACTNAME
# define ACTNAME  "yacc.act"
#endif

/* output file name */
#ifndef OFILE
# define OFILE "y_tab.sh"
#endif

/* user output file name */
#ifndef FILEU
# define FILEU "y_output"
#endif

/* command to clobber tempfiles after use */
#ifndef ZAPFILE
# define ZAPFILE(x) unlink(x)
#endif

static void error(const char *f, ...) __attribute__ ((format (printf, 1, 2)));

# define Y2_IDENTIFIER 257
# define Y2_MARK 258
# define Y2_TERM 259
# define Y2_LEFT 260
# define Y2_RIGHT 261
# define Y2_BINARY 262
# define Y2_PREC 263
# define Y2_LCURLY 264
# define Y2_C_IDENTIFIER 265  /* name followed by colon */
# define Y2_NUMBER 266
# define Y2_START 267
# define Y2_ENDFILE 0


static void finact(void)
{
	/* finish action routine */
	fclose(faction);
	fprintf( ftable, "\ndeclare -i YYERRCODE=%d\n", tokset[2].value );
}

static char *cstash( char *s )
{
	char *temp;

	temp = cnamp;
	do {
		if( cnamp >= &cnames[cnamsz] ) error(ediag(
						"too many characters in names and literals",
						"      "
						) );
		else *cnamp++ = *s;
		}  while ( *s++ );
	return( temp );
}

static int defin( int t, char  *s )
{
/*      define s to be a terminal if t=0
	or a nonterminal if t=1         */

	int val = 0;

	if (t) {
		if( ++nnonter >= NNONTERM ) error(ediag(
						"too many non-terminals, max %d",
						"  , . %d"
						), NNONTERM);
		nontrst[nnonter].name = cstash(s);
		return( NTBASE + nnonter );
		}
	/* must be a token */
	if( ++ntokens >= NTERMS ) error(ediag(
					"too many tarminals, max %d",
					"  , . %d"
					), NTERMS );
	tokset[ntokens].name = cstash(s);

	/* establish value for token */

	if( s[0]==' ' && s[2]=='\0' ) /* single character literal */
		val = s[1];
	else if ( s[0]==' ' && s[1]=='\\' ) { /* escape sequence */
		if( s[3] == '\0' ){ /* single character escape sequence */
			switch ( s[2] ){
					 /* character which is escaped */
			case 'n': val = '\n'; break;
			case 'r': val = '\r'; break;
			case 'b': val = '\b'; break;
			case 't': val = '\t'; break;
			case 'f': val = '\f'; break;
			case '\'': val = '\''; break;
			case '"': val = '"'; break;
			case '\\': val = '\\'; break;
			default: error(ediag("invalid \\-escaping", " \\-"));
				}
			}
		else if( s[2] <= '7' && s[2]>='0' ){ /* \nnn sequence */
			if( s[3]<'0' || s[3] > '7' || s[4]<'0' ||
				s[4]>'7' || s[5] != '\0' ) error(ediag(
								"invalid \\nnn-escaping",
								"  \\nnn"));
			val = 64*s[2] + 8*s[3] + s[4] - 73*'0';
			if( val == 0 ) error(ediag("unsupport '\\000'", "'\\000' "));
			}
		}
	else {
		val = extval++;
		}
	tokset[ntokens].value = val;
	toklev[ntokens] = 0;
	return( ntokens );
}

static void defout(void) { /* write out the defines (at the end of the declaration section) */

	int i, c, n = 0;
	char *cp;

	for( i=ndefout; i<=ntokens; ++i ){

		cp = tokset[i].name;
		if( *cp == ' ' ) ++cp;  /* literals */

		for( ; (c= *cp)!='\0'; ++cp ){

			if( islower(c) || isupper(c) || isdigit(c) || c=='_' );  /* VOID */
			else goto nodef;
			}

		if( n == 0 )
			fprintf( ftable, "declare -i");
		fprintf( ftable, " %s=%d", tokset[i].name, tokset[i].value );
		n++;
		if (n == 4) {
			putc('\n', ftable);
			n = 0;
		}

	nodef:  ;
	}

	if( n )
		putc('\n', ftable);
	ndefout = ntokens+1;

}


static int skipcom(void) { /* skip over comments */
	int c, i=0;  /* i is the number of lines skipped */

	/* skipcom is called after reading a / */

	if( getc(finput) != '*' ) error(ediag("invalid a remark", " "));
	c = getc(finput);
	while( c != EOF ){
		while( c == '*' ){
			if( (c=getc(finput)) == '/' ) return( i );
			}
		if( c == '\n' ) ++i;
		c = getc(finput);
		}
	error(ediag("EOF in a remark", "EOF  "));
	/* NOTREACHED */
	return 0;
}

static int gettok(void) {
	int i, base;
	static int peekline; /* number of '\n' seen in lookahead */
	int c, match, reserve;

begin:
	reserve = 0;
	lineno += peekline;
	peekline = 0;
	c = getc(finput);
	while( c==' ' || c=='\n' || c=='\t' || c=='\f' ){
		if( c == '\n' ) ++lineno;
		c=getc(finput);
		}
	if( c == '/' ){ /* skip comment */
		lineno += skipcom();
		goto begin;
		}

	switch(c) {

	case EOF:
		return(Y2_ENDFILE);
	case '{':  /* '}' */
		ungetc( c, finput );
		return( '=' );  /* action ... */
	case '"':
	case '\'':
		match = c;
		tokname[0] = ' ';
		i = 1;
		for(;;){
			c = getc(finput);
			if( c == '\n' || c == EOF )
				error(ediag("invalid or unclosed %c",
					"    %c"), match);
			if( c == '\\' ){
				c = getc(finput);
				tokname[i] = '\\';
				if( ++i >= NAMESIZE ) --i;
				}
			else if( c == match ) break;
			tokname[i] = c;
			if( ++i >= NAMESIZE ) --i;
			}
		break;

	case '%':
	case '\\':

		switch(c=getc(finput)) {

		case '0':       return(Y2_TERM);
		case '<':       return(Y2_LEFT);
		case '2':       return(Y2_BINARY);
		case '>':       return(Y2_RIGHT);
		case '%':
		case '\\':      return(Y2_MARK);
		case '=':       return(Y2_PREC);
		case '{':       return(Y2_LCURLY); /* '}' */
		default:        reserve = 1;
			}

	default:

		if( isdigit(c) ){ /* number */
			numbval = c-'0' ;
			base = (c=='0') ? 8 : 10 ;
			for( c=getc(finput); isdigit(c) ; c=getc(finput) ){
				numbval = numbval*base + c - '0';
				}
			ungetc( c, finput );
			return(Y2_NUMBER);
			}
		else if( islower(c) || isupper(c) || c=='_' || c=='.' || c=='$' ){
			i = 0;
			while( islower(c) || isupper(c) || isdigit(c) || c=='_' || c=='.' || c=='$' ){
				tokname[i] = c;
				if( reserve && isupper(c) ) tokname[i] += 'a'-'A';
				if( ++i >= NAMESIZE ) --i;
				c = getc(finput);
				}
			}
		else return(c);

		ungetc( c, finput );
		}

	tokname[i] = '\0';

	if( reserve ){ /* find a reserved word */
		if( !strcmp(tokname,"term")) return( Y2_TERM );
		if( !strcmp(tokname,"token")) return( Y2_TERM );
		if( !strcmp(tokname,"left")) return( Y2_LEFT );
		if( !strcmp(tokname,"nonassoc")) return( Y2_BINARY );
		if( !strcmp(tokname,"binary")) return( Y2_BINARY );
		if( !strcmp(tokname,"right")) return( Y2_RIGHT );
		if( !strcmp(tokname,"prec")) return( Y2_PREC );
		if( !strcmp(tokname,"start")) return( Y2_START );
		error(ediag("the keyword '%s' unknown", "  : %s"), tokname );
		}

	/* look ahead to distinguish Y2_IDENTIFIER from Y2_C_IDENTIFIER */

	c = getc(finput);
	while( c==' ' || c=='\t'|| c=='\n' || c=='\f' || c== '/' ) {
		if( c == '\n' ) ++peekline;
		else if( c == '/' ){ /* look for comments */
			peekline += skipcom();
			}
		c = getc(finput);
		}
	if( c == ':' ) return( Y2_C_IDENTIFIER );
	ungetc( c, finput );
	return( Y2_IDENTIFIER );
}

static int chfind( int t, char *s ) {
	int i;

	if (s[0]==' ') t=0;
	TLOOP(i){
		if(!strcmp(s,tokset[i].name)){
			return( i );
			}
		}
	NTLOOP(i){
		if(!strcmp(s,nontrst[i].name)) {
			return( i+NTBASE );
			}
		}
	/* cannot find name */
	if( t>1 )
		error(ediag("'%s' undefined", "%s "), s);
	return( defin( t, s ) );
}

static void cpycode(void) { /* copies code between \{ and \} */

	int c;
	c = getc(finput);
	if( c == '\n' ) {
		c = getc(finput);
		lineno++;
		}
	fprintf( ftable, "\n#line %d \"%s\"\n", lineno, infile );
	while( c>=0 ){
		if( c== '\\' ) {
			if( (c=getc(finput)) == '}' ) return;
			else putc('\\', ftable );
		}
		if( c== '%' ) {
			if( (c=getc(finput)) == '}' ) return;
			else putc('%', ftable );
		}
		putc( c , ftable );
		if( c == '\n' ) ++lineno;
		c = getc(finput);
		}
	error(ediag("EOF before %%", "EOF  %%}"));
}

/* copy C action to the next ; or closing } */
static void cpyact(int offset)
{
	int brac = 0, c, match, j, s;
	fpos_t pos;

loop:
	c = getc(finput);
swt:
	switch( c ) {

	case ';':
		if( brac == 0 )
			return;
		if( brac > 1)
			goto lcopy;
		/* skip tail ";[[:spaces:]]*}" */
		fgetpos(finput, &pos);
		while (1) {
			c = getc(finput);
			if( c == EOF )
				goto swt;
			if( c == '}' )
				return;
			if( !isspace(c) ) {
				fsetpos(finput, &pos);
				c = ';';
				goto lcopy;
			}
		}

	case '\n':
		++lineno;
		c = getc(finput);
		if( c == EOF )
			goto swt;
		if( c != '\t' && c != ' ')
			ungetc( c, finput );
		c = '\n';
		if( brac > 1 )
			goto lcopy;
		/* skip tail "\n[[:spaces:]]*[;}]" */
		fgetpos(finput, &pos);
		match = brac == 0 ? ';' : '}';
		while (1) {
			c = getc(finput);
			if( c == EOF )
				goto swt;
			if( c == match ) {
				putc( ' ' , faction );
				return;
			}
			if( !isspace(c) ) {
				fsetpos(finput, &pos);
				c = '\n';
				goto lcopy;
			}
		}
	case '{':
		brac++;
		if( brac > 1 )
			putc( c , faction );
		goto loop;

	case '@':
		c = getc(finput);
		if( c == '"' || c == '\'' )
			goto lcopy;
		s = 1;
		if( c == '@' ) {
			c = getc(finput);
			if(c == '=') {
				fprintf( faction, "yyval=");
			} else {
				ungetc( c, finput );
				fprintf( faction, "$yyval");
			}
			goto loop;
		}
		if( c == '-' ){
			s = -s;
			c = getc(finput);
		}
		if( isdigit(c) ){
			j=0;
			while( isdigit(c) ){
				j= j*10+c-'0';
				c = getc(finput);
				}

			j = j*s - offset;
			if( j > 0 )
				error(ediag("invalid @%d usage", "  @%d"), j+offset );

			if( j != 0 )
				fprintf( faction, "${yyv[yypvt-%d]}", -j );
			else
				fprintf( faction, "${yyv[yypvt]}" );
			goto swt;
		}
		putc( '@' , faction );
		if( s<0 ) putc('-', faction );
		goto swt;

	case '}':
		if( --brac ) goto lcopy;
		return;


	case '/':       /* look for comments */
		c = getc(finput);
		if( c != '*' ) {
			putc( '/' , faction );
			goto swt;
		}

		/* it really is a comment */
		putc( ' ' , faction );
		putc( '#' , faction );
		c = getc(finput);
		while( c != EOF ){
			if( c == '*' ) {
				c = getc(finput);
				if( c == '/' ) {
					c = '\n';
					goto lcopy;
				}
				putc( '*' , faction );
				if( c == '\n' ) ++lineno;
				continue;
			} else
				putc( c , faction );
			if( c == '\n' )++lineno;
			c = getc(finput);
		}
		error(ediag("EOF in a remark", "EOF  "));

	case '\'':      /* character constant */
		match = '\'';
		goto string;

	case '"':       /* character string */
		match = '"';

	string:

		putc( c , faction );
		while( (c=getc(finput)) != EOF ){

			if( c=='\\' ){
				putc( c , faction );
				c=getc(finput);
				if( c == '\n' ) ++lineno;
				}
			else if( c==match ) goto lcopy;
			else if( c=='\n' ) error(ediag("EOL in a character or a string",
						"EOL     "));
			putc( c , faction );
			}
		error(ediag("EOF in a character or a string", "EOF     "));

	case EOF:
		error(ediag("unclosed a rule", "  "));

	}

lcopy:
	putc( c , faction );
	goto loop;
}

#include <getopt.h>

static void setup(int argc, char *argv[])
{       int i,j,lev,t, ty;
	int c;
	int *p;
	char actname[8];
	char *ofile = OFILE;
	int verbose = 0;

	foutput = NULL;
	i = 1;

	while ((c = getopt(argc, argv, "viso:")) != -1) {
		switch(c) {
			case 'v':
				verbose++;
				break;
			case 'i':
				YYSTYPE="-i ";
				break;
			case 's':
				YYSTYPE="";
				break;
			case 'o':
				ofile = optarg;
				break;
			default:
				error(ediag("the option is unknown", "  "));
		}
	}
	argc -= optind;
	argv += optind;
	if( verbose ) {
		foutput = fopen(FILEU, "w" );
		if( foutput == NULL ) error(ediag("can not create '%s'", "   %s"), FILEU);
	}
	if( ofile[0] == '\0' || (ofile[0] == '-' && ofile[0] == '\0'))
		ftable = stdout;
	else {
		ftable = fopen( ofile, "w" );
		if( ftable == NULL ) error(ediag("can not create '%s'","   %s"), ofile );
	}
	fprintf( ftable,  "#!/usr/bin/env bash\n\n" );
	fprintf( ftable,  "# yychar: a current lexem, yyerrflag: then flag of recovery after error\n" );
	fprintf( ftable,  "declare -i yychar yyerrflag\n" );
	fprintf( ftable,  "declare %syylval yyval\n", YYSTYPE );
	ftemp = fopen( TEMPNAME, "w" );
	faction = fopen( ACTNAME, "w" );
	if( ftemp==NULL || faction==NULL ) error(ediag("can not create temporary files",
						"    "));
	if( (cnamp = cnames = malloc(CNAMSZ)) == NULL ) {
		fprintf(stderr,"No mem to cnames\n");
		exit(2);
	}
	if( argc != 1 )
		error(ediag("can not process more one files", "    "));
	if( (finput=fopen( infile=argv[0], "r" )) == NULL )
		error(ediag("can not open '%s' input file", "     %s"), infile);

	defin(0,"$end");
	extval = 0400;
	defin(0,"error");
	defin(1,"$accept");
	mem=mem0;
	lev = 0;
	ty = 0;
	i=0;

	/* sorry -- no yacc parser here... we must bootstrap somehow... */

	for( t=gettok();  t!=Y2_MARK && t!= Y2_ENDFILE; ){
		switch( t ){

		case ';':
			t = gettok();
			break;

		case Y2_START:
			if( (t=gettok()) != Y2_IDENTIFIER ) {
				error("bad %%start" );
			}
			start = chfind(1,tokname);
			t = gettok();
			continue;

		case Y2_LEFT:
		case Y2_BINARY:
		case Y2_RIGHT:
			++i;
		case Y2_TERM:
			lev = t-Y2_TERM;  /* nonzero means new prec. and assoc. */
			ty = 0;

			/* get identifiers so defined */

			t = gettok();

			for(;;) {
				switch( t ){

				case ',':
					t = gettok();
					continue;

				case ';':
					break;

				case Y2_IDENTIFIER:
					j = chfind(0,tokname);
					if( lev ){
						if( ASSOC(toklev[j]) )
							error(ediag("redefined previous set '%s'",
								"  %s"), tokname);
						SETASC(toklev[j],lev);
						SETPLEV(toklev[j],i);
						}
					if( ty ){
						if( TYPE(toklev[j]) )
							error(ediag("redifined the type '%s'",
								"  %s"), tokname );
						SETTYPE(toklev[j],ty);
						}
					if( (t=gettok()) == Y2_NUMBER ){
						tokset[j].value = numbval;
						if( j < ndefout && j>2 ){
							error(ediag("undefined the type '%s'",
								" %s   "), tokset[j].name);
							}
						t=gettok();
						}
					continue;

					}

				break;
				}

			continue;

		case Y2_LCURLY:
			defout();
			cpycode();
			t = gettok();
			continue;

		default:
			error(ediag("syntax error", " "));
			}

		}

	if( t == Y2_ENDFILE ) {
		error(ediag("EOF, can not find %%", "  ( %%)"));
	}

	/* t is Y2_MARK */

	defout();

	fprintf( ftable,  "\nyyclearin() { yychar=-1; }\n" );
	fprintf( ftable,  "yyerrok() { yyerrflag=0; }\n" );

	prdptr[0]=mem;
	/* added production */
	*mem++ = NTBASE;
	*mem++ = start;  /* if start is 0, we will overwrite with the lhs of the first rule */
	*mem++ = 1;
	*mem++ = 0;
	prdptr[1]=mem;

	while( (t=gettok()) == Y2_LCURLY ) cpycode();

	if( t != Y2_C_IDENTIFIER ) error(ediag("syntax error in the first rule",
						"    "));

	if( !start ) prdptr[0][1] = chfind(1,tokname);

	/* read rules */

	while( t!=Y2_MARK && t!=Y2_ENDFILE ){

		/* process a rule */

		if( t == '|' ){
			*mem++ = *prdptr[nprod-1];
			}
		else if( t == Y2_C_IDENTIFIER ){
			*mem = chfind(1,tokname);
			if( *mem < NTBASE ) error(ediag("the terminal in the left part a rule",
							"    "));
			++mem;
			}
		else error(ediag("invalid the rule, may be forgotted ':' or '|'",
					" :  :  | ?"));

		/* read rule body */


		t = gettok();
	more_rule:
		while( t == Y2_IDENTIFIER ) {
			*mem = chfind(1,tokname);
			if( *mem<NTBASE ) levprd[nprod] = toklev[*mem];
			++mem;
			t = gettok();
			}


		if( t == Y2_PREC ){
			if( gettok()!=Y2_IDENTIFIER) error(ediag("invalid the %%prec syntax", " %%prec"));
			j = chfind(2,tokname);
			if( j>=NTBASE)error(ediag("disalow a non-terminal after %%prec",
							" %s   %%prec"),
							nontrst[j-NTBASE].name);
			levprd[nprod]=toklev[j];
			t = gettok();
			}

		if( t == '=' ){
			levprd[nprod] |= ACTFLAG;
			fprintf( faction, "\n\t\t %d)", nprod );
			cpyact( mem-prdptr[nprod]-1 );
			fprintf( faction, " ;;" );
			if( (t=gettok()) == Y2_IDENTIFIER ){
				/* action within rule... */

				sprintf( actname, "$$%d", nprod );
				j = chfind(1,actname);  /* make it a nonterminal */

				/* the current rule will become rule number nprod+1 */
				/* move the contents down, and make room for the null */

				for( p=mem; p>=prdptr[nprod]; --p ) p[2] = *p;
				mem += 2;

				/* enter null production for action */

				p = prdptr[nprod];

				*p++ = j;
				*p++ = -nprod;

				/* update the production information */

				levprd[nprod+1] = levprd[nprod] & ~ACTFLAG;
				levprd[nprod] = ACTFLAG;

				if( ++nprod >= NPROD ) error(ediag("the rules counter overflow, max %d",
								",  %d "), NPROD );
				prdptr[nprod] = p;

				/* make the action appear in the original rule */
				*mem++ = j;

				/* get some more of the rule */

				goto more_rule;
				}
			}

		while( t == ';' ) t = gettok();

		*mem++ = -nprod;

		/* check that default action is reasonable */

		if( ++nprod >= NPROD ) error(ediag("the rules counter overflow, max %d",
								",  %d "), NPROD );
		prdptr[nprod] = mem;
		levprd[nprod]=0;

		}

	/* end of all rules */

	finact();
	if( t == Y2_MARK )
		Y2_MARK_fi = finput;
	else
		fclose( finput );
}

static void gin(int i) {

	int *p, *r, *s, *q1, *q2;

	/* enter gotos on nonterminal i into array a */

	ggreed[i] = 0;

	q2 = mem0+ yypgo[i+1] - 1;
	q1 = mem0 + yypgo[i];

	/* now, find a place for it */

	for( p=amem; p < &amem[ACTSIZE]; ++p ){
		if( *p ) continue;
		for( r=q1; r<q2; r+=2 ){
			s = p + *r +1;
			if( *s ) goto nextgp;
			if( s > maxa ){
				if( (maxa=s) > &amem[ACTSIZE] ) error(ediag("overflow of the 'amem' dimension",
							"  amem"));
				}
			}
		/* we have found a spot */

		*p = *q2;
		if( p > maxa ){
			if( (maxa=p) > &amem[ACTSIZE] ) error(ediag("internal overflow of the 'amem' dimension",
							"  amem"));
			}
		for( r=q1; r<q2; r+=2 ){
			s = p + *r + 1;
			*s = r[1];
			}

		pgo[i] = p-amem;
		goto nextgi;

		nextgp:  ;
		}

	error(ediag("can not goto %d", "  goto %d"), i);

	nextgi:  ;
}

static void stin(int i) {
	int *r, *s, n, flag, j, *q1, *q2;

	tystate[i] = 0;

	/* enter state i into the a array */

	q2 = mem0+temp1[i+1];
	q1 = mem0+temp1[i];
	/* find an acceptable place */

	for( n= -maxoff; n<ACTSIZE; ++n ){

		flag = 0;
		for( r = q1; r < q2; r += 2 ){
			if( (s = *r + n + amem ) < amem ) goto nextn;
			if( *s == 0 ) ++flag;
			else if( *s != r[1] ) goto nextn;
			}

		/* check that the position equals another only if the states are identical */

		for( j=0; j<nstate; ++j ){
			if( indgo[j] == n ) {
				if( flag ) goto nextn;  /* we have some disagreement */
				if( temp1[j+1] + temp1[i] == temp1[j] + temp1[i+1] ){
					/* states are equal */
					indgo[i] = n;
					return;
					}
				goto nextn;  /* we have some disagreement */
				}
			}

		for( r = q1; r < q2; r += 2 ){
			if( (s = *r + n + amem ) >= &amem[ACTSIZE] ) error(ediag("overflow of the 'amem' dimension",
							"  amem"));
			if( s > maxa ) maxa = s;
			if( *s != 0 && *s != r[1] ) error(ediag(
						"the 'amem' dimension corrupted, a #%td position, a element is %d",
						" amem, . no %td, . %d"), s-amem, r[1] );
			*s = r[1];
			}
		indgo[i] = n;
		return;

		nextn:  ;
		}

	error(ediag("the error state %d", "  %d"), i);
}

static int nxti(void) { /* finds the next i */
	int i, max, maxi = -1;

	max = 0;

	for( i=1; i<= nnonter; ++i ) if( ggreed[i] >= max ){
		max = ggreed[i];
		maxi = -i;
		}

	for( i=0; i<nstate; ++i ) if( tystate[i] >= max ){
		max = tystate[i];
		maxi = i;
		}

	if( max==0 ) return( Y4_NOMORE );
	else return( maxi );
}

static void osummary(void) {
	/* write summary */

	int i, *p;

	if( foutput == NULL ) return;
	i=0;
	for( p=maxa; p>=amem; --p ) {
		if( *p == 0 ) ++i;
		}

	fprintf( foutput, ediag("Optimizer space used: input %td/%d, output %td/%d\n",
				" :  %td/%d,  %td/%d\n"),
		pmem-mem0+1, MEMSIZE, maxa-amem+1, ACTSIZE );
	fprintf( foutput, ediag("%td table entries, %d zero\n",
				"%td  , %d \n"),
	       (maxa-amem)+1, i );
	fprintf( foutput, ediag("maximum spread: %d, maximum offset: %d\n",
				". spread: %d, . : %d\n"),
	       maxspr, maxoff );

}

static void arout( char *s, int *v, int n ) {

	int i;

	fprintf( ftable, "\ndeclare -i -a %s=( \n", s );
	for( i=0; i<n; ) {
		if( i!=0 && i%10 == 0 ) fprintf( ftable, "\n" );
		fprintf( ftable, "%4d", v[i] );
		if( ++i == n ) fprintf( ftable, " )\n" );
		else fprintf( ftable, " " );
	}
}

static void aoutput(void)  /* this version is for C */
{
	/* write out the optimized parser */

	fprintf( ftable, "declare -i YYLAST=%td\n", maxa-amem+1 );

	arout( "yyact", amem, (maxa-amem)+1 );
	arout( "yypact", indgo, nstate );
	arout( "yypgo", pgo, nnonter+1 );
}



static int gtnm(void) {

	int s, val, c;

	/* read and convert an integer from the standard input */
	/* return the terminating character */
	/* blanks, tabs, and newlines are ignored */

	s = 1;
	val = 0;

	while( (c=getc(finput)) != EOF ){
		if( isdigit(c) ){
			val = val * 10 + c - '0';
			}
		else if ( c == '-' ) s = -1;
		else break;
		}

	*pmem++ = s*val;
	if( pmem > &mem0[MEMSIZE] ) error(ediag("exhausted a internal memory", "   "));
	return( c );

	}

static void callopt(void)
{
	int i, *p, j, k, *q;
	pmem = mem0;
	/* read the arrays from tempfile and set parameters */

	if( (finput=fopen(TEMPNAME,"r")) == NULL )
		 error(ediag("can not reopen temporary file", " a %s ??"), TEMPNAME );
	pgo[0] = 0;
	temp1[0] = 0;
	nstate = 0;
	nnonter = 0;
	for(;;){
		switch( gtnm() ){

		case '\n':
			temp1[++nstate] = (--pmem) - mem0;
		case ',':
			continue;

		case '$':
			break;

		default:
			error(ediag("the temporary file corrupted", "  "));
			}
		break;
		}

	temp1[nstate] = yypgo[0] = (--pmem) - mem0;

	for(;;){
		switch( gtnm() ){

		case '\n':
			yypgo[++nnonter]= pmem-mem0;
		case ',':
			continue;

		case EOF:
			break;

		default:
			error(ediag("the temporary file corrupted", "  "));
			}
		break;
		}

	yypgo[nnonter--] = (--pmem) - mem0;



	for( i=0; i<nstate; ++i ){

		k = 32000;
		j = 0;
		q = mem0 + temp1[i+1];
		for( p = mem0 + temp1[i]; p<q ; p += 2 ){
			if( *p > j ) j = *p;
			if( *p < k ) k = *p;
			}
		if( k <= j ){ /* nontrivial situation */
			/* temporarily, kill this for compatibility j -= k; */
			/* j is now the range */
			if( k > maxoff ) maxoff = k;
			}
		tystate[i] = (temp1[i+1]-temp1[i]) + 2*j;
		if( j > maxspr ) maxspr = j;
		}

	/* initialize ggreed table */

	for( i=1; i<=nnonter; ++i ){
		ggreed[i] = 1;
		j = 0;
		/* minimum entry index is always 0 */
		q = mem0 + yypgo[i+1] -1;
		for( p = mem0+yypgo[i]; p<q ; p += 2 ) {
			ggreed[i] += 2;
			if( *p > j ) j = *p;
			}
		ggreed[i] = ggreed[i] + 2*j;
		if( j > maxoff ) maxoff = j;
		}


	/* now, prepare to put the shift actions into the amem array */

	for( i=0; i<ACTSIZE; ++i ) amem[i] = 0;
	maxa = amem;

	for( i=0; i<nstate; ++i ) {
		indgo[i] = YYFLAG1;
		}

	while( (i = nxti()) != Y4_NOMORE ) {
		if( i >= 0 ) stin(i);
		else gin(-i);

		}

	/* write out the output appropriate to the language */

	aoutput();

	osummary();
	ZAPFILE(TEMPNAME);
}

static void summary(void) /* output the summary on the tty */
{
	/*zzmemsz = mem0; */
	if( foutput!=NULL ){
		fprintf( foutput, ediag("\n%d/%d terminals, %d/%d nonterminals\n",
					"\n%d/%d , %d/%d \n"),
			 ntokens, NTERMS, nnonter, NNONTERM );
		fprintf( foutput, ediag("%d/%d grammar rules, %d/%d states\n",
					"%d/%d , %d/%d \n"),
			 nprod, NPROD, nstate, NSTATES );
		fprintf( foutput, ediag("%d shift/reduce, %d reduce/reduce conflicts reported\n",
					": %d /, %d /\n"),
			 zzsrconf, zzrrconf );
		fprintf( foutput, ediag("%td/%d working sets used\n",
					"%td/%d  \n"),
			 zzcwp-wsets,  WSETSIZE );
		fprintf( foutput, ediag("memory: states,etc. %td/%d, parser %td/%d\n",
					":   .. %td/%d,  %td/%d\n"),
			 zzmemsz-mem0, MEMSIZE, memp-amem, ACTSIZE );
		fprintf( foutput, ediag("%d/%d distinct lookahead sets\n",
					"%d/%d  -.\n"),
			 nlset, LSETSIZE );
		fprintf( foutput, ediag("%d extra closures\n",
					"%d \n"), zzclose - 2*nstate );
		fprintf( foutput, ediag("%d shift entries, %d exceptions\n",
					"%d , %d \n"), zzacent, zzexcp );
		fprintf( foutput, ediag("%d goto entries\n",
					"%d \n"), zzgoent );
		fprintf( foutput, ediag("%d entries saved by goto default\n",
					"%d   \n"), zzgobest );
		fclose( foutput );
		}
	if( zzsrconf!=0 || zzrrconf!=0 ) {
		printf(ediag("\nconflicts: ", "\n: "));
		if( zzsrconf ) printf(ediag("%d shift/reduce", "%d /"), zzsrconf);
		if( zzsrconf && zzrrconf )printf(", " );
		if( zzrrconf ) printf(ediag("%d reduce/reduce", "%d /"), zzrrconf );
		printf("\n" );
	}
}

/* VARARGS2 */
void error(const char *f, ...) /* write out error comment */
{
	va_list p;

	va_start(p, f);
	++nerrors;
	fprintf( stderr, ediag("\nfatal error: ", "\n: "));
	vfprintf( stderr, f, p);
	va_end(p);
	fprintf( stderr, ediag(", line %d\n", ", . %d\n"), lineno );
	if( !fatfl ) return;
	summary();
	exit(1);
}


static void warray(char *s, int *v, int n )
{
	int i;

	fprintf( ftable, "\ndeclare -i -a %s=(\n", s );
	for( i=0; i<n; ){
		if( i!= 0 && i%10 == 0 ) fprintf( ftable, "\n" );
		fprintf( ftable, "%4d", v[i] );
		if( ++i == n ) fprintf( ftable, " )\n" );
		else fprintf( ftable, " " );
		}
}


static char *chcopy( char *p, char *q ) {
	/* copies string q into p, returning next free char ptr */
	while( (*p = *q++)!='\0' ) ++p;
	return( p );
}

static char *symnam(int i) /* return a pointer to the name of symbol i */
{
	char *cp;

	cp = (i>=NTBASE) ? nontrst[i-NTBASE].name : tokset[i].name ;
	if( *cp == ' ' ) ++cp;
	return( cp );
}



static void aryfil( int *v, int n, int c ) /* set elements 0 through n-1 to c */
{
	int i;

	for( i=0; i<n; ++i ) v[i] = c;
}

static int setunion( int *a, int *b )
{
	/* set a to the union of a and b */
	/* return 1 if b is not a subset of a, 0 otherwise */
	int i, x, sub = 0;

	SETLOOP(i) {
		*a = (x = *a)|*b++;
		if( *a++ != x ) sub = 1;
	}
	return( sub );
}

/* compute an array with the beginnings of  productions yielding given nonterminals
   The array pres points to these lists */
/* the array pyield has the lists: the total size is only NPROD+1 */
static void cpres(void)
{
	int **pmem;
	int c, j, i;
	static int * pyield[NPROD];

	pmem = pyield;

	NTLOOP(i) {
		c = i+NTBASE;
		pres[i] = pmem;
		fatfl = 0;  /* make undefined  symbols  nonfatal */
		PLOOP(0,j) {
			if (*prdptr[j] == c) *pmem++ =  prdptr[j]+1;
		}
		if(pres[i] == pmem)
			error(ediag("can not find the '%s' non-terminal", "  %s"), nontrst[i].name);
	}
	pres[i] = pmem;
	fatfl = 1;
	if( nerrors ) {
		summary();
		exit(1);
	}
	if( pmem != &pyield[nprod] )
		error(ediag("the internal yacc error, pyield %td", "  Yacc': pyield %td"), pmem-&pyield[nprod]);
}

/* decide if the lookahead set pointed to by p is known */
/* return pointer to a perminent location for the set */
static struct looksets *flset( struct looksets *p )
{
	struct looksets *q;
	int j, *w;
	int *u, *v;

	for( q = &lkst[nlset]; q-- > lkst; ){
		u = p->lset;
		v = q->lset;
		w = & v[tbitset];
		while( v<w) if( *u++ != *v++ ) goto more;
		/* we have matched */
		return( q );
		more: ;
	}
	/* add a new one */
	q = &lkst[nlset++];
	if( nlset >= LSETSIZE )error(ediag("exhausted internal memory", "    -"));
	SETLOOP(j) {
		q->lset[j] = p->lset[j];
	}
	return( q );
}

static void cpfir(void) {
	/* compute an array with the first of nonterminals */
	int *p, **s, i, **t, ch, changes;

	zzcwp = &wsets[nnonter];
	NTLOOP(i){
		aryfil( wsets[i].ws.lset, tbitset, 0 );
		t = pres[i+1];
		for( s=pres[i]; s<t; ++s ){ /* initially fill the sets */
			for( p = *s; (ch = *p) > 0 ; ++p ) {
				if( ch < NTBASE ) {
					SETBIT( wsets[i].ws.lset, ch );
					break;
					}
				else if( !pempty[ch-NTBASE] ) break;
				}
			}
		}

	/* now, reflect transitivity */

	changes = 1;
	while( changes ){
		changes = 0;
		NTLOOP(i){
			t = pres[i+1];
			for( s=pres[i]; s<t; ++s ){
				for( p = *s; ( ch = (*p-NTBASE) ) >= 0; ++p ) {
					changes |= setunion( wsets[i].ws.lset, wsets[ch].ws.lset );
					if( !pempty[ch] ) break;
					}
				}
			}
		}

	NTLOOP(i) pfirst[i] = flset( &wsets[i].ws );
}

/* sorts last state,and sees if it equals earlier ones. returns state number */
static int state(int c)
{
	int size1,size2;
	int i;
	struct item *p1, *p2, *k, *l, *q1, *q2;

	p1 = pstate[nstate];
	p2 = pstate[nstate+1];
	if(p1 == p2) return(0); /* null state */

	/* sort the items */
	for(k=p2-1;k>p1;k--) {  /* make k the biggest */
		for(l=k-1;l>=p1;--l)if( l->pitem > k->pitem ){
			int *s;
			struct looksets *ss;
			s = k->pitem;
			k->pitem = l->pitem;
			l->pitem = s;
			ss = k->look;
			k->look = l->look;
			l->look = ss;
		}
	}
	size1 = p2 - p1; /* size of state */

	for( i= (c>=NTBASE)?ntstates[c-NTBASE]:tstates[c]; i != 0; i = mstates[i] ) {
		/* get ith state */
		q1 = pstate[i];
		q2 = pstate[i+1];
		size2 = q2 - q1;
		if (size1 != size2) continue;
		k=p1;
		for(l=q1;l<q2;l++) {
			if( l->pitem != k->pitem ) break;
			++k;
		}
		if (l != q2) continue;
		/* found it */
		pstate[nstate+1] = pstate[nstate]; /* delete last state */
		/* fix up lookaheads */
		if( nolook ) return(i);
		for( l=q1,k=p1; l<q2; ++l,++k ){
			int s;
			SETLOOP(s) clset.lset[s] = l->look->lset[s];
			if( setunion( clset.lset, k->look->lset ) ) {
				tystate[i] = MUSTDO;
				/* register the new set */
				l->look = flset( &clset );
				}
		}
		return i;
	}

	/* state is new */
	if( nolook ) error(ediag("internal yacc error in the state/nolook"," yacc' state/nolook"));
	pstate[nstate+2] = p2;
	if(nstate+1 >= NSTATES) error(ediag("too many states", "  "));
	if( c >= NTBASE ) {
		mstates[ nstate ] = ntstates[ c-NTBASE ];
		ntstates[ c-NTBASE ] = nstate;
	} else {
		mstates[ nstate ] = tstates[ c ];
		tstates[ c ] = nstate;
	}
	tystate[nstate]=MUSTDO;
	return(nstate++);
}

static void putitem( int *ptr, struct looksets *lptr )
{
	struct item *j;

	j = pstate[nstate+1];
	j->pitem = ptr;
	if( !nolook ) j->look = flset( lptr );
	pstate[nstate+1] = ++j;
	if( (int *)j > zzmemsz ) {
		zzmemsz = (int *)j;
		if( zzmemsz >=  &mem0[MEMSIZE] ) error(ediag("exhausted internal memory for states",
								"   "));
	}
}

static void cempty(void) { /* mark nonterminals which derive the empty string */
	/* also, look for nonterminals which don't derive any token strings */

# define Y1_EMPTY 1
# define Y1_WHOKNOWS 0
# define Y1_OK 1

	int i, *p;

	/* first, use the array pempty to detect productions that can never be reduced */
	/* set pempty to WHONOWS */
	aryfil( pempty, nnonter+1, Y1_WHOKNOWS );

	/* now, look at productions, marking nonterminals which derive something */

	more:
	PLOOP(0,i) {
		if( pempty[ *prdptr[i] - NTBASE ] ) continue;
		for( p=prdptr[i]+1; *p>=0; ++p ) {
			if( *p>=NTBASE && pempty[ *p-NTBASE ] == Y1_WHOKNOWS ) break;
		}
		if( *p < 0 ) { /* production can be derived */
			pempty[ *prdptr[i]-NTBASE ] = Y1_OK;
			goto more;
		}
	}

	/* now, look at the nonterminals, to see if they are all Y1_OK */

	NTLOOP(i) {
		/* the added production rises or falls as the start symbol ... */
		if( i == 0 ) continue;
		if( pempty[ i ] != Y1_OK ) {
			fatfl = 0;
			error(ediag("the non-terminal '%s' have empty producing",
					" %s   "), nontrst[i].name );
		}
	}

	if( nerrors ) {
		summary();
		exit(1);
	}

	/* now, compute the pempty array, to see which nonterminals derive the empty string */

	/* set pempty to Y1_WHOKNOWS */

	aryfil( pempty, nnonter+1, Y1_WHOKNOWS );

	/* loop as long as we keep finding empty nonterminals */

again:
	PLOOP(1,i) {
		if( pempty[ *prdptr[i]-NTBASE ]==Y1_WHOKNOWS ) { /* not known to be empty */
			for( p=prdptr[i]+1; *p>=NTBASE && pempty[*p-NTBASE]==Y1_EMPTY ; ++p ) ;
			if( *p < 0 ) { /* we have a nontrivially empty nonterminal */
				pempty[*prdptr[i]-NTBASE] = Y1_EMPTY;
				goto again; /* got one ... try for another */
			}
		}
	}

}

static void closure(int i) { /* generate the closure of state i */

	int c, ch, work, k;
	struct wset *u, *v;
	int *pi;
	int **s, **t;
	struct item *q;
	struct item *p;

	++zzclose;

	/* first, copy kernel of state i to wsets */

	cwp = wsets;
	ITMLOOP(i,p,q){
		cwp->pitem = p->pitem;
		cwp->flag = 1;    /* this item must get closed */
		SETLOOP(k) cwp->ws.lset[k] = p->look->lset[k];
		WSBUMP(cwp);
		}

	/* now, go through the loop, closing each item */

	work = 1;
	while( work ){
		work = 0;
		WSLOOP(wsets,u){

			if( u->flag == 0 ) continue;
			c = *(u->pitem);  /* dot is before c */

			if( c < NTBASE ){
				u->flag = 0;
				continue;  /* only interesting case is where . is before nonterminal */
				}

			/* compute the lookahead */
			aryfil( clset.lset, tbitset, 0 );

			/* find items involving c */

			WSLOOP(u,v){
				if( v->flag == 1 && *(pi=v->pitem) == c ){
					v->flag = 0;
					if( nolook ) continue;
					while( (ch= *++pi)>0 ){
						if( ch < NTBASE ){ /* terminal symbol */
							SETBIT( clset.lset, ch );
							break;
							}
						/* nonterminal symbol */
						setunion( clset.lset, pfirst[ch-NTBASE]->lset );
						if( !pempty[ch-NTBASE] ) break;
						}
					if( ch<=0 ) setunion( clset.lset, v->ws.lset );
					}
				}

			/*  now loop over productions derived from c */

			c -= NTBASE; /* c is now nonterminal number */

			t = pres[c+1];
			for( s=pres[c]; s<t; ++s ){
				/* put these items into the closure */
				WSLOOP(wsets,v){ /* is the item there */
					if( v->pitem == *s ){ /* yes, it is there */
						if( nolook ) goto nexts;
						if( setunion( v->ws.lset, clset.lset ) ) v->flag = work = 1;
						goto nexts;
						}
					}

				/*  not there; make a new entry */
				if( cwp-wsets+1 >= WSETSIZE )
				    error(ediag("exhausted a internal work memory", "  "));
				cwp->pitem = *s;
				cwp->flag = 1;
				if( !nolook ){
					work = 1;
					SETLOOP(k) cwp->ws.lset[k] = clset.lset[k];
					}
				WSBUMP(cwp);

			nexts: ;
				}

			}
		}

	/* have computed closure; flags are reset; return */
	if( cwp > zzcwp ) zzcwp = cwp;
}


static int apack(int *p, int n ) { /* pack state i from temp1 into amem */
	int off;
	int *pp, *qq, *rr;
	int *q, *r;

	/* we don't need to worry about checking because we
	   we will only look entries known to be there... */

	/* eliminate leading and trailing 0's */

	q = p+n;
	for( pp=p,off=0 ; *pp==0 && pp<=q; ++pp,--off ) /* VOID */ ;
	if( pp > q ) return(0);  /* no actions */
	p = pp;

	/* now, find a place for the elements from p to q, inclusive */

	r = &amem[ACTSIZE-1];
	for( rr=amem; rr<=r; ++rr,++off ){  /* try rr */
		for( qq=rr,pp=p ; pp<=q ; ++pp,++qq){
			if( *pp != 0 ){
				if( *pp != *qq && *qq != 0 ) goto nextk;
				}
			}

		/* we have found an acceptable k */

		for( qq=rr,pp=p; pp<=q; ++pp,++qq ){
			if( *pp ){
				if( qq > r ) error(ediag("the actions table overflow", "  "));
				if( qq>memp ) memp = qq;
				*qq = *pp;
				}
			}
		return off;

		nextk: ;
	}
	error(ediag("the actions table overflow", "  "));
	/* NOTREACHED */
	return 0;
}


static void go2gen(int c) { /* output the gotos for nonterminal c */

	int i, work, cc;
	struct item *p, *q;


	/* first, find nonterminals with gotos on c */

	aryfil( temp1, nnonter+1, 0 );
	temp1[c] = 1;

	work = 1;
	while( work ){
		work = 0;
		PLOOP(0,i){
			if( (cc=prdptr[i][1]-NTBASE) >= 0 ){ /* cc is a nonterminal */
				if( temp1[cc] != 0 ){ /* cc has a goto on c */
					cc = *prdptr[i]-NTBASE; /* thus, the left side of production i does too */
					if( temp1[cc] == 0 ){
						  work = 1;
						  temp1[cc] = 1;
						  }
					}
				}
			}
		}

	/* now, we have temp1[c] = 1 if a goto on c in closure of cc */

	/* now, go through and put gotos into tystate */

	aryfil( tystate, nstate, 0 );
	SLOOP(i){
		ITMLOOP(i,p,q){
			if( (cc= *p->pitem) >= NTBASE ){
				if( temp1[cc -= NTBASE] ){ /* goto on c is possible */
					tystate[i] = amem[indgo[i]+c];
					break;
					}
				}
			}
		}
}

static void precftn(int r, int t, int s) /* decide a shift/reduce conflict by precedence. */
{
	/* r is a rule number, t a token number *
	 * the conflict is in state s *
	 * temp1[t] is changed to reflect the action */

	int lp,lt, action;

	lp = levprd[r];
	lt = toklev[t];
	if( PLEVEL(lt) == 0 || PLEVEL(lp) == 0 ) {
		/* conflict */
		if( foutput != NULL ) fprintf( foutput, ediag(
						"\n%d: shift/reduce conflict (shift %d, red'n %d) on %s",
						"\n%d:  / ( %d,  %d)  %s"),
						s, temp1[t], r, symnam(t) );
		++zzsrconf;
		return;
		}
	if( PLEVEL(lt) == PLEVEL(lp) ) action = ASSOC(lt);
	else if( PLEVEL(lt) > PLEVEL(lp) ) action = RASC;  /* shift */
	else action = LASC;  /* reduce */

	switch( action ){

	case BASC:  /* error action */
		temp1[t] = ERRCODE;
		return;

	case LASC:  /* reduce */
		temp1[t] = -r;
		return;

		}
	}


# define Y1_ISIZE 400
static char *writem(int *pp)   /* creates output string for item pointed to by pp */
{
	int i,*p;
	static char sarr[Y1_ISIZE];
	char *q;

	for( p=pp; *p>0 ; ++p ) ;
	p = prdptr[-*p];
	q = chcopy( sarr, nontrst[*p-NTBASE].name );
	q = chcopy( q, " : " );

	for(;;){
		*q++ = ++p==pp ? '_' : ' ';
		*q = '\0';
		if((i = *p) <= 0) break;
		q = chcopy( q, symnam(i) );
		if( q> &sarr[Y1_ISIZE-30] ) error(ediag("the element is too long", "  "));
		}

	if( (i = *pp) < 0 ){ /* an item calling for a reduction */
		q = chcopy( q, "    (" );
		sprintf( q, "%d)", -i );
		}

	return( sarr );
}

static void wrstate(int i) { /* writes state i */
	int j0,j1;
	struct item *pp, *qq;
	struct wset *u;

	if( foutput == NULL ) return;
	fprintf( foutput, ediag( "\nstate %d\n","\n %d\n"),i);
	ITMLOOP(i,pp,qq) fprintf( foutput, "\t%s\n", writem(pp->pitem));
	if( tystate[i] == MUSTLOOKAHEAD ){
		/* print out empty productions in closure */
		WSLOOP( wsets+(pstate[i+1]-pstate[i]), u ){
			if( *(u->pitem) < 0 ) fprintf( foutput, "\t%s\n", writem(u->pitem) );
			}
		}

	/* check for state equal to another */

	TLOOP(j0) if( (j1=temp1[j0]) != 0 ){
		fprintf( foutput, "\n\t%s\t", symnam(j0) );
		if( j1>0 ){ /* shift, error, or accept */
			if( j1 == ACCEPTCODE ) fprintf( foutput,  "accept" );
			else if( j1 == ERRCODE ) fprintf( foutput, "error" );
			else fprintf( foutput, ediag("shift %d", " %d"), j1 );
			}
		else fprintf( foutput, ediag("reduce %d", "\t%d"),-j1 );
		}

	/* output the final production */

	if( lastred ) fprintf( foutput, ediag(
				"\n\t.  reduce %d\n\n",
				"\n\t.   %d\n\n"), lastred );
	else fprintf( foutput, "\n\t.  error\n\n" );

	/* now, output nonterminal actions */

	j1 = ntokens;
	for( j0 = 1; j0 <= nnonter; ++j0 ) {
		if( temp1[++j1] ) fprintf( foutput, ediag(
					   "\t%s  goto %d\n",
					   "\t%s\t\t%d\n"), symnam( j0+NTBASE), temp1[j1] );
	}

}

static void wdef( char *s, int n ) { /* output a definition of s to the value n */
	fprintf( ftable, "declare -i %s=%d\n", s, n );
}


static void hideprod(void)
{
	/* in order to free up the mem and amem arrays for the optimizer,
	 * and still be able to output yyr1, etc., after the sizes of
	 * the action array is known, we hide the nonterminals
	 * derived by productions in levprd. */

	int i, j;

	j = 0;
	levprd[0] = 0;
	PLOOP(1,i){
		if( !(levprd[i] & REDFLAG) ){
			++j;
			if( foutput != NULL ){
				fprintf( foutput, ediag(
						"Rule not reduced:   %s\n",
						"  : %s\n"), writem( prdptr[i] ) );
				}
			}
		levprd[i] = *prdptr[i] - NTBASE;
		}
	if( j ) printf( ediag(
				"%d rules never reduced\n",
				"%d   \n"), j );
}

static void wract(int i) { /* output state i */
	/* temp1 has the actions, lastred the default */
	int p, p0, p1;
	int ntimes, tred, count, j;
	int flag;

	/* find the best choice for lastred */

	lastred = 0;
	ntimes = 0;
	TLOOP(j){
		if( temp1[j] >= 0 ) continue;
		if( temp1[j]+lastred == 0 ) continue;
		/* count the number of appearances of temp1[j] */
		count = 0;
		tred = -temp1[j];
		levprd[tred] |= REDFLAG;
		TLOOP(p){
			if( temp1[p]+tred == 0 ) ++count;
			}
		if( count >ntimes ){
			lastred = tred;
			ntimes = count;
			}
		}

	/* for error recovery, arrange that, if there is a shift on the
	 * error recovery token, `error', that the default be the error action */
	if( temp1[1] > 0 ) lastred = 0;

	/* clear out entries in temp1 which equal lastred */
	TLOOP(p) if( temp1[p]+lastred == 0 )temp1[p]=0;

	wrstate(i);
	defact[i] = lastred;

	flag = 0;
	TLOOP(p0){
		if( (p1=temp1[p0])!=0 ) {
			if( p1 < 0 ){
				p1 = -p1;
				goto exc;
				}
			else if( p1 == ACCEPTCODE ) {
				p1 = -1;
				goto exc;
				}
			else if( p1 == ERRCODE ) {
				p1 = 0;
				goto exc;
			exc:
				if( flag++ == 0 ) fprintf( ftable, "\t-1 %d\n", i );
				fprintf( ftable, "\t%d %d\n", tokset[p0].value, p1 );
				++zzexcp;
				}
			else {
				fprintf( ftemp, "%d,%d,", tokset[p0].value, p1 );
				++zzacent;
				}
			}
		}
	if( flag ) {
		defact[i] = -2;
		fprintf( ftable, "\t-2 %d\n", lastred );
		}
	fprintf( ftemp, "\n" );
}

static void output(void) { /* print the output for the states */

	int i, k, c;
	struct wset *u, *v;

	fprintf( ftable, "declare -i -a yyexca=(\n" );

	SLOOP(i) { /* output the stuff for state i */
		nolook = !(tystate[i]==MUSTLOOKAHEAD);
		closure(i);
		/* output actions */
		nolook = 1;
		aryfil( temp1, ntokens+nnonter+1, 0 );
		WSLOOP(wsets,u){
			c = *( u->pitem );
			if( c>1 && c<NTBASE && temp1[c]==0 ) {
				WSLOOP(u,v){
					if( c == *(v->pitem) ) putitem( v->pitem+1, (struct looksets *)0 );
					}
				temp1[c] = state(c);
				}
			else if( c > NTBASE && temp1[ (c -= NTBASE) + ntokens ] == 0 ){
				temp1[ c+ntokens ] = amem[indgo[i]+c];
				}
			}

		if( i == 1 ) temp1[1] = ACCEPTCODE;

		/* now, we have the shifts; look at the reductions */

		lastred = 0;
		WSLOOP(wsets,u){
			c = *( u->pitem );
			if( c<=0 ){ /* reduction */
				lastred = -c;
				TLOOP(k){
					if( BIT(u->ws.lset,k) ){
						  if( temp1[k] == 0 ) temp1[k] = c;
						  else if( temp1[k]<0 ){ /* reduce/reduce conflict */
						    if( foutput!=NULL )
						      fprintf( foutput, ediag(
							"\n%d: reduce/reduce conflict (red'ns %d and %d ) on %s",
							"\n%d:  / ( %d  %d)  %s"),
							i, -temp1[k], lastred, symnam(k) );
						    if( -temp1[k] > lastred ) temp1[k] = -lastred;
						    ++zzrrconf;
						    }
						  else { /* potential shift/reduce conflict */
							precftn( lastred, k, i );
							}
						}
					}
				}
			}
		wract(i);
		}

	fprintf( ftable, ")\n\n" );

	wdef( "YYNPROD", nprod );

}

static void go2out(void) { /* output the gotos for the nontermninals */
	int i, j, k, best, count, cbest, times;

	fprintf( ftemp, "$\n" );  /* mark begining of gotos */

	for( i=1; i<=nnonter; ++i ) {
		go2gen(i);

		/* find the best one to make default */

		best = -1;
		times = 0;

		for( j=0; j<=nstate; ++j ){ /* is j the most frequent */
			if( tystate[j] == 0 ) continue;
			if( tystate[j] == best ) continue;

			/* is tystate[j] the most frequent */

			count = 0;
			cbest = tystate[j];

			for( k=j; k<=nstate; ++k ) if( tystate[k]==cbest ) ++count;

			if( count > times ){
				best = cbest;
				times = count;
				}
			}

		/* best is now the default entry */

		zzgobest += (times-1);
		for( j=0; j<=nstate; ++j ){
			if( tystate[j] != 0 && tystate[j]!=best ){
				fprintf( ftemp, "%d,%d,", j, tystate[j] );
				zzgoent += 1;
				}
			}

		/* now, the default */

		zzgoent += 1;
		fprintf( ftemp, "%d\n", best );

		}
}

static void stagen(void) { /* generate the states */

	int i, c;
	struct wset *p, *q;

	/* initialize */

	nstate = 0;
	/* THIS IS FUNNY from the standpoint of portability
	 * it represents the magic moment when the mem0 array, which has
	 * been holding the productions, starts to hold item pointers, of a
	 * different type...
	 * someday, alloc should be used to allocate all this stuff... for now, we
	 * accept that if pointers don't fit in integers, there is a problem... */

	pstate[0] = pstate[1] = (struct item *)mem;
	aryfil( clset.lset, tbitset, 0 );
	putitem( prdptr[0]+1, &clset );
	tystate[0] = MUSTDO;
	nstate = 1;
	pstate[2] = pstate[1];

	aryfil( amem, ACTSIZE, 0 );

	/* now, the main state generation loop */

	more:
	SLOOP(i) {
		if( tystate[i] != MUSTDO ) continue;
		tystate[i] = DONE;
		aryfil( temp1, nnonter+1, 0 );
		/* take state i, close it, and do gotos */
		closure(i);
		WSLOOP(wsets,p) { /* generate goto's */
			if( p->flag ) continue;
			p->flag = 1;
			c = *(p->pitem);
			if( c <= 1 ) {
				if( pstate[i+1]-pstate[i] <= p-wsets ) tystate[i] = MUSTLOOKAHEAD;
				continue;
			}
			/* do a goto on c */
			WSLOOP(p,q) {
				if( c == *(q->pitem) ) { /* this item contributes to the goto */
					putitem( q->pitem + 1, &q->ws );
					q->flag = 1;
				}
			}
			if( c < NTBASE ) {
				state(c);  /* register new state */
			} else {
				temp1[c-NTBASE] = state(c);
			}
		}
		indgo[i] = apack( &temp1[1], nnonter-1 ) - 1;
		goto more; /* we have done one goto; do some more */
	}
	/* no more to do... stop */
}

static unsigned char Parser[] = "\
\n\
declare -i YYFLAG=-1000\n\
# return values:\n\
# 0 - YYACCEPT\n\
# 1 - YYABORT\n\
\n\
yyparse() {\n\
	local -i -a yys\n\
	local @T-a yyv # the stack for values saving\n\
	local -i yyj yym yyn yystate=0 yypvt yyps=-1 yypv=-1 ns=0\n\
\n\
	yychar=-1\n\
	yyerrflag=0\n\
\n\
	# yystack: push state to the stack\n\
	while true; do\n\
		if [[ ns -eq 0 ]]; then\n\
			yys[++yyps]=yystate\n\
			yyv[++yypv]=$yyval\n\
		else\n\
			ns=0\n\
		fi\n\
\n\
		# yynewstate:\n\
		yyn=yypact[yystate]\n\
		if [[ yyn -gt YYFLAG ]]; then\n\
			# complex state\n\
			[[ yychar -lt 0 ]] && yylex\n\
			yyn+=yychar\n\
			if [[ yyn -ge 0 && yyn -lt YYLAST ]]; then\n\
				yyn=yyact[yyn]\n\
				if [[ yychk[yyn] -eq yychar ]]; then # may be shift\n\
					yychar=-1\n\
					yyval=$yylval\n\
					yystate=yyn\n\
					[[ yyerrflag -gt 0 ]] && yyerrflag=yyerrflag-1\n\
					continue # goto yystack\n\
				fi\n\
			fi\n\
		fi\n\
\n\
		# default action\n\
\n\
		yyn=yydef[yystate]\n\
		if [[ yyn -eq -2 ]]; then\n\
			[[ yychar -lt 0 ]] && yylex\n\
\n\
			# find from the exeption table\n\
			for((ns=0; (yyexca[ns]!=(-1)) || (yyexca[ns+1]!=yystate); ns+=2)); do\n\
				true\n\
			done\n\
\n\
			while true; do\n\
				ns+=2\n\
				[[ yyexca[ns] -lt 0 ]] && break\n\
				[[ yyexca[ns] -eq yychar ]] && break\n\
			done\n\
			yyn=yyexca[ns+1]\n\
			[[ yyn -lt 0 ]] && return 0   # OK\n\
			ns=0\n\
		fi\n\
\n\
		if [[ yyn -eq 0 ]]; then\n\
			# error ... try continue\n\
\n\
			case $yyerrflag in\n\
			 0)     # found first error\n\
				# yyerror \"syntax error\"\n\
				return 1;;\n\
\n\
			 1|2)   # incomplete recovery, try again\n\
				yyerrflag=3\n\
\n\
				# find a state for allow shifting if error\n\
\n\
				while [[ yyps -ge 0 ]]; do\n\
					yyn=yypact[yys[yyps]]+YYERRCODE\n\
					if [[ yyn -ge 0 && yyn -lt YYLAST && yychk[yyact[yyn]] -eq YYERRCODE ]]; then\n\
						# imitation shift while error\n\
						yystate=yyact[yyn]\n\
						continue 2 # goto yystack\n\
					fi\n\
					yyn=yypact[yys[yyps]]\n\
\n\
					# in current 'yyps' do not found shifting for error, popup from the stack\n\
\n\
					yyps=yyps-1\n\
					yypv=yypv-1\n\
				done\n\
\n\
				# in stack do not found a state of shifting... abort\n\
				return 1;;\n\
\n\
			 3)     # cannot shifting, get then lexem\n\
				[[ yychar -eq 0 ]] && return 1 # EOF, abort\n\
				yychar=-1\n\
				# goto yynewstate: try again with a current state\n\
				ns=1\n\
				continue;;\n\
			esac\n\
		fi\n\
\n\
		# reduce 'yyn' rule\n\
\n\
		yyps=yyps-yyr2[yyn];\n\
		yypvt=yypv\n\
		yypv=yypv-yyr2[yyn]\n\
		yyval=${yyv[yypv+1]}\n\
		yym=yyn\n\
\n\
		# get a next state from the table of states\n\
		yyn=yyr1[yyn]\n\
		yyj=yypgo[yyn]+yys[yyps]+1\n\
		if [[ yyj -ge YYLAST ]]; then\n\
			yystate=yyact[yypgo[yyn]]\n\
		else\n\
			yystate=yyact[yyj]\n\
			[[ yychk[yystate] -ne -yyn ]] && yystate=yyact[yypgo[yyn]]\n\
		fi\n\
		case $yym in@@\n\
		esac\n\
	done  # push a new state and the value to stack\n\
}\n";

/* put out other arrays, copy the parsers */
static void others(void)
{
	int c, i, j;
	unsigned char *pp = Parser;

	warray( "yyr1", levprd, nprod );

	aryfil( temp1, nprod, 0 );
	PLOOP(1,i)temp1[i] = prdptr[i+1]-prdptr[i]-2;
	warray( "yyr2", temp1, nprod );

	aryfil( temp1, nstate, -1000 );
	TLOOP(i){
		for( j=tstates[i]; j!=0; j=mstates[j] ){
			temp1[j] = tokset[i].value;
			}
		}
	NTLOOP(i){
		for( j=ntstates[i]; j!=0; j=mstates[j] ){
			temp1[j] = -i;
			}
		}
	warray( "yychk", temp1, nstate );

	warray( "yydef", defact, nstate );

	/* copy parser text */
	while( (c = *pp ) != 0 ) {
		if( c != '@' )
			putc( c, ftable );
		else {
			c = *++pp;
			if( c == 'T' ) {
				fputs( YYSTYPE, ftable );
			} else if( c == '@' ) {
				/* copy actions */
				faction = fopen( ACTNAME, "r" );
				if( faction == NULL ) error(ediag("can not open '%s'.", "    %s."),
							ACTNAME );
				while( (c=getc(faction) ) != EOF ) putc( c, ftable );
				fclose(faction);
				ZAPFILE(ACTNAME);
			} else {
				putc( '@', ftable );
				continue;
			}
		}
		pp++;
	}

	if(Y2_MARK_fi) {
		fprintf( ftable, "\n# line %d \"%s\"\n", lineno, infile );
		while( (c=getc(Y2_MARK_fi)) != EOF ) putc( c, ftable );
	}

	fclose( ftable );
}

int main(int argc, char *argv[])
{
	setbuf(stdout,NULL);
	setbuf(stderr,NULL);
	setbuf(stdin,NULL);

	amem = memp = malloc(ACTSIZE*sizeof(int));
	zzmemsz = mem0;
	if( amem == NULL ) {
err_mem:
		fprintf(stderr, ediag("Not enoug memory\n","  \n"));
		exit(1);
	}
	mem0 = malloc(MEMSIZE*sizeof(int));
	if( mem0 == NULL ) goto err_mem;
	temp1 = malloc(TEMPSIZE*sizeof(int));
	if( temp1 == NULL ) goto err_mem;

	setup(argc, argv); /* initialize and read productions */
	if(foutput)
		fprintf( stderr, ediag("YACC 2.1 to bash\nReading\n", "YACC 2.1  bash\n\n") );
	tbitset = NWORDS(ntokens);

	if(foutput)
		fprintf( stderr, ediag("Tables\n", "T\n") );
	cpres(); /* make table of which productions yield a given nonterminal */
	cempty(); /* make a table of which nonterminals can match the empty string */
	cpfir(); /* make a table of firsts of nonterminals */

	if(foutput)
		fprintf( stderr, ediag("States\n","\n") );
	stagen(); /* generate the states */

	if(foutput)
		fprintf( stderr, ediag("Output\n","\n") );
	output();  /* write the states and the tables */
	go2out();
	hideprod();
	fclose( ftemp ); /* Bug fixed by @VG: 13.05.85 22:18 */
	summary();

	if(foutput)
		fprintf( stderr, ediag("Optimizing\n","\n") );
	callopt();

	if(foutput)
		fprintf( stderr, ediag("Output\n","\n") );
	others();

	if(foutput)
		fprintf( stderr, ediag("Ready\n","\n") );
	return 0;
}
